# utils/rgb_composite.py

# External Imports
import numpy as np

# Local Imports
from .normalize_image import normalize_image


def make_rgb_composite(
    r: np.ndarray,
    g: np.ndarray,
    b: np.ndarray,
    low_percentile: int = 5,
    high_percentile: int = 95,
):
    """
    Creates an RGB false color composite image from three image-like arrays.
    """
    rdata = r[np.isfinite(r)]
    gdata = g[np.isfinite(g)]
    bdata = b[np.isfinite(b)]

    rnorm = normalize_image(
        r,
        float(np.percentile(rdata, low_percentile)),
        float(np.percentile(rdata, high_percentile)),
    )
    gnorm = normalize_image(
        g,
        float(np.percentile(gdata, low_percentile)),
        float(np.percentile(gdata, high_percentile)),
    )
    bnorm = normalize_image(
        b,
        float(np.percentile(bdata, low_percentile)),
        float(np.percentile(bdata, high_percentile)),
    )

    rgb_composite = np.concat(
        [
            rnorm[:, :, np.newaxis],
            gnorm[:, :, np.newaxis],
            bnorm[:, :, np.newaxis],
        ],
        axis=2,
    )

    return rgb_composite



# Standard Libraries
from typing import Optional

# Dependencies
import numpy as np
import numpy.typing as npt


class SingleFitResult:
    def __init__(
        self,
        xdata: npt.NDArray,
        ydata: npt.NDArray,
        model: npt.NDArray,
        beta: npt.NDArray,
        res: npt.NDArray,
    ):
        self.xdata = xdata
        self.ydata = ydata
        self.model = model
        self.beta = beta
        self.res = res

    def r_squared(self) -> npt.NDArray[np.float32]:
        """
        Returns the coefficient of determination (R<sup>2</sup>)
        """
        ss_res = np.sum(self.res**2)
        ss_tot = np.sum((self.ydata - np.mean(self.ydata)) ** 2)
        return 1 - (ss_res / ss_tot)

    def eval(self, x: npt.NDArray):
        """
        Returns dependent variable value given independent variable.
        """
        X = np.vander(x, self.beta.size, increasing=True)
        return (X @ self.beta[:, None])[:, 0]


def polyfit_single(
    spectrum: np.ndarray,
    wvl: np.ndarray,
    order: int = 1,
    design_matrices: Optional[tuple[np.ndarray, ...]] = None,
    return_coefficients: bool = False,
) -> SingleFitResult:
    """
    Fits a polynomial of order `N` to `x` and `y` data. Optionally can be used
    in a loop by specifying the `design_matrix` elements.

    Parameters
    ----------
    spectrum: np.ndarray
        Y Data. Spectral Data.
    wvl: np.ndarray
        X Data. Wavelengths. Can be None if `design_matrices` are supplied.
    order: int
        Order of polynomial fit. Can be None if `design_matrices` are supplied.
    design_matrices: Nothing or tuple[np.ndarray]
        Three design matrix components. X, Xt and XtX. If None (default),
        these are calculated from `wvl` data.
    return_coefficients: bool, optional
        If True (defualt), returns fit coefficients rather than a fit line.

    Returns
    -------
    beta: np.ndarray
        Coefficients of the fit.
    """

    if design_matrices is None:
        X = np.vander(wvl, order + 1, increasing=True)
        Xt = X.T
        XtX = Xt @ X
    else:
        X, Xt, XtX = design_matrices

    beta = np.linalg.inv(XtX) @ (Xt @ spectrum)
    model = X @ beta
    residual = model - spectrum

    return SingleFitResult(wvl, spectrum, model, beta, residual)



# Dependencies
import numpy as np
import numpy.typing as npt
from scipy.signal import convolve

# Relative Imports
from .polyfit import polyfit_single


def box_filter_single(
    spectrum: np.ndarray, window_size: int = 5
) -> tuple[npt.NDArray, npt.NDArray]:
    """
    Smoothes a spectrum using a moving average.

    Parameters
    ----------
    original_spectrum: np.ndarray
        Non-smooth spectrum.
    window_size: optional, int
        Window size to use for the moving average. Default is 5.
    edge_handling: optional, str
        How to handle the edge cases of spectrum after convolution. Default is
        `"extrapolate"`
    remove_outliers: optional, bool
        If true, outliers are removed from the spectrum. A threshold value
        can be passed. Default is False.
    outlier_threshold: optional, float
        Sigma threshold to be used in outlier removal. See
        `outlier_removal.py`. Default is 2.
    """
    spec: npt.NDArray = spectrum

    window = np.ones(window_size)
    nbands = spec.size

    edge_length = int(np.maximum(round(spec.size * 0.1, 0), 2))
    left_idx: npt.NDArray[np.int32] = np.arange(
        0, 1 + edge_length, dtype=np.int32
    )
    right_idx: npt.NDArray[np.int32] = np.arange(
        nbands - edge_length, nbands, dtype=np.int32
    )

    left_fit = polyfit_single(spec[left_idx], left_idx, 1)
    left_ext = np.arange(-window_size, 0)
    left_arm = left_fit.eval(left_ext)

    right_fit = polyfit_single(spec[right_idx], right_idx, 1)
    right_ext = np.arange(nbands, nbands + window_size)
    right_arm = right_fit.eval(right_ext)

    smoothed_spectrum = np.empty(nbands + left_arm.size + right_arm.size)

    smoothed_spectrum[: left_arm.size] = left_arm
    smoothed_spectrum[left_arm.size : -right_arm.size] = spec  # noqa
    smoothed_spectrum[-right_arm.size :] = right_arm  # noqa

    mu = convolve(smoothed_spectrum, window, mode="same") / window_size
    musq = convolve(smoothed_spectrum**2, window, mode="same") / window_size
    sigma = np.sqrt(musq - mu**2)

    mu = mu[window_size:-window_size]
    sigma = sigma[window_size:-window_size]

    return mu, sigma



# Dependencies
import numpy as np

# Relative Imports
from .utils import CubeInterpolator

# Top-Level Imports
from reflspeckit.data_classes import Wavelength


def double_line(
    cube: np.ndarray, wvls: Wavelength
) -> tuple[np.ndarray, np.ndarray]:
    """
    Double-line continuum removal.

    Parameters
    ----------
    spectrum: np.ndarray
        Filtered spectrum cube with the continuum in place.
    wvls: Wavelength
        Wavelength values in the form of a Wavelength object.

    Returns
    -------
    continuum_removed: np.ndarray
        Input spectrum cube with the continuum removed.
    continuum: np.ndarray
        Continuum values for the spectrum.
    """
    # Getting initial continuum line parameters
    anchor_pts = np.array([700, 1550, 2600])
    if wvls.unit == "nm":
        pass
    if wvls.unit == "um":
        anchor_pts = anchor_pts * 10**-3
    if wvls.unit == "m":
        anchor_pts = anchor_pts * 10**-9

    cont1_band_idx = np.empty(anchor_pts.size, dtype=np.int16)
    cont1_band_wvls = np.empty(anchor_pts.size, dtype=np.float64)
    for n in range(anchor_pts.size):
        idx = np.argmin(np.abs(wvls.values - anchor_pts[n]))
        cont1_band_idx[n] = idx
        cont1_band_wvls[n] = wvls.values[idx]

    cont1_spectrum_values = cube[:, :, cont1_band_idx]

    CI1 = CubeInterpolator(cont1_band_wvls, cont1_spectrum_values)
    continuum1 = CI1.linear(wvls.values)

    continuum1_removed = cube / continuum1

    cont2_ranges: dict[str, tuple[float, float]] = {
        "range1": (650.0, 1000.0),
        "range2": (1350.0, 1600.0),
        "range3": (2000.0, 2600.0),
    }

    def scale_dict_values(
        data: dict[str, tuple[float, float]], factor: float
    ) -> dict[str, tuple[float, float]]:
        return {k: (v[0] * factor, v[1] * factor) for k, v in data.items()}

    if wvls.unit == "nm":
        pass
    elif wvls.unit == "um":
        cont2_ranges = scale_dict_values(cont2_ranges, 10**-3)
    elif wvls.unit == "m":
        cont2_ranges = scale_dict_values(cont2_ranges, 10**-9)
    else:
        raise ValueError("Wavelength Units not valid.")

    cont2_band_idx = np.empty(
        (*cube.shape[:2], len(cont2_ranges)), dtype=np.int16
    )

    # Locating dynamic maxima for second continuum parameters
    for n, val in enumerate(cont2_ranges.values()):
        lo_idx = np.argmin(np.abs(wvls.values - val[0]))
        hi_idx = np.argmin(np.abs(wvls.values - val[1]))

        max_spectrum_idx = (
            np.argmax(
                continuum1_removed[
                    :, :, np.arange(lo_idx, hi_idx, dtype=np.int16)
                ],
                axis=2,
            )
            + lo_idx
        )
        cont2_band_idx[:, :, n] = max_spectrum_idx

    cont2_band_wvl = wvls.values[cont2_band_idx]
    _rows = np.arange(cube.shape[0])[:, None, None]
    _cols = np.arange(cube.shape[1])[None, :, None]
    cont2_spectrum_values = cube[_rows, _cols, cont2_band_idx]

    CI2 = CubeInterpolator(cont2_band_wvl, cont2_spectrum_values)
    continuum2 = CI2.linear(wvls.values)
    continuum2_removed = cube / continuum2

    return continuum2_removed, continuum2


def single_line(spectrum: np.ndarray, wvls: np.ndarray, tie_points: tuple):
    """
    Numba-optimized version of `single_line`.

    Parameters
    ----------
    wvls: np.ndarray
        Wavelength values of the spectrum (in nm).
    spectrum: np.ndarray
        Spectrum values.
    tie_points: tuple of floats
        Tie points used to interpolate continuum.

    Returns
    -------
    continuum_removed: np.ndarray
        Spectrum with the continuum removed.
    continuum: np.ndarray
        The continuum values.
    """
    # Getting initial continuum line parameters
    anchor_pts = np.array([700, 1550, 2600])

    cont1_band_idx = np.empty(anchor_pts.size, dtype=np.int16)
    cont1_band_wvls = np.empty(anchor_pts.size, dtype=np.float64)
    for n in range(anchor_pts.size):
        idx = np.argmin(np.abs(wvls - anchor_pts[n]))
        cont1_band_idx[n] = idx
        cont1_band_wvls[n] = wvls[idx]

    cont1_spectrum_values = spectrum[cont1_band_idx]

    CI1 = CubeInterpolator(cont1_band_wvls, cont1_spectrum_values)
    continuum1 = CI1.linear(wvls)

    continuum1_removed = spectrum / continuum1

    return continuum1_removed, continuum1



# Dependencies
import numpy as np
import numpy.typing as npt


class CubeFitResult:
    def __init__(
        self,
        xdata: npt.NDArray,
        ydata: npt.NDArray,
        model: npt.NDArray,
        beta: npt.NDArray,
        res: npt.NDArray,
    ):
        self.xdata = xdata
        self.ydata = ydata
        self.model = model
        self.beta = beta
        self.res = res

    def r_squared(self) -> npt.NDArray[np.float32]:
        """
        Returns the coefficient of determination (R<sup>2</sup>)
        """
        ss_res = np.sum(self.res**2, axis=2)
        ss_tot = np.sum(
            (self.ydata - np.mean(self.ydata, axis=-1)[:, :, None]) ** 2
        )
        return 1 - (ss_res / ss_tot)

    def eval(self, x: float | npt.NDArray):
        """
        Returns dependent variable value given independent variable.
        """
        return self.beta @ x


def polyfit_cube(
    spectral_cube: np.ndarray, wvl: np.ndarray, order: int
) -> CubeFitResult:
    """
    Performs polynomial fits for an entire spectral cube of data.

    Parameters
    ----------
    spectral_cube: np.ndarray
        Spectral data cube.
    wvl: np.ndarray
        Wavelength values.
    order: int
        Order of polyfit.
    return_coefficients: bool, optional
        If True (defualt), returns a cube of coefficient rather than fit lines.

    Returns
    -------
    fit_cube: np.ndarray
        Either a cube of fitted coefficients or fitted lines.
    """
    X = np.vander(wvl, order + 1)
    prefix = np.linalg.inv(X.T @ X) @ X.T
    beta = np.einsum("ij,...j->...i", prefix, spectral_cube)
    model = np.einsum("ij,...j->...i", X, beta)
    residual = spectral_cube - model

    return CubeFitResult(wvl, spectral_cube, model, beta, residual)



import numpy as np


def last_nonzero_val_cube(cube, return_index=False):
    """
    If you have an empty 3D image array with the first two dimensions being
    pixels and the third dimension of size N, and each pixel is filled in to a
    certain depth, M <= N, this function returns a 2D image array that picks
    out all the pixel values at position M.
    """
    nx, ny, _ = cube.shape
    result = np.full((nx, ny), np.nan, dtype=cube.dtype)
    for i in range(nx):
        for j in range(ny):
            vals = cube[i, j, :]
            nonzero_indices = np.nonzero(np.isfinite(vals))[0]
            if return_index:
                result[i, j] = (
                    nonzero_indices[-1] if nonzero_indices.size > 0 else np.nan
                )
            else:
                result[i, j] = (
                    vals[nonzero_indices[-1]]
                    if nonzero_indices.size > 0
                    else np.nan
                )
    return result



# Built-Ins
from copy import copy

# Dependencies
import numpy as np
import numpy.typing as npt

# Top-Level Imports
from reflspeckit.data_classes import (
    Wavelength,
    WvlUnit,
    FilterMethodLiteral,
    ContinuumMethodLiteral,
)
from reflspeckit._errors import DimensionError

# Relative Imports
from .filtering import box_filter_single
from .outlier_detection import remove_outliers
from .continuum_removal import double_line, single_line
from .absorption_feature1d import AbsorptionFeature1D


class Spec1D:
    """
    Main class for handling 1-dimensional spectral data (i.e. a single
    spectrum).

    Parameters
    ----------
    spec_arr: np.ndarray
    wvl_arr: np.ndarray
        Wavelength values of the spectrum. Must be a 1D array.
    unit: WvlUnit, optional
        Unit for the wavelength values. Default is "nm", but options include:

        - "nm": Nanometers
        - "um": Microns (micrometers)
        - "m": Meters

    Notes
    -----
    See Spec3D for an equivalent class that handles 3-dimensional spectral data
    image cubes.
    """

    def __init__(
        self, spec_arr: npt.NDArray, wvl_arr: npt.NDArray, unit: WvlUnit = "nm"
    ):
        self.spectrum = spec_arr
        self.wavelength = Wavelength(wvl_arr, "nm")

        self.original_spectrum = copy(spec_arr)
        self.no_outliers: npt.NDArray = np.full_like(spec_arr, np.nan)
        self.filtered: npt.NDArray = np.full_like(spec_arr, np.nan)
        self.noise: npt.NDArray = np.full_like(spec_arr, np.nan)
        self.contrem: npt.NDArray = np.full_like(spec_arr, np.nan)
        self.continuum: npt.NDArray = np.full_like(spec_arr, np.nan)

        self._validate()

    def outlier_removal(self, sigma_threshold: float = 1.5):
        no_outliers = remove_outliers(self.spectrum, sigma_threshold)
        self.no_outliers = no_outliers
        self.spectrum = no_outliers

    def noise_reduction(
        self,
        method: FilterMethodLiteral,
        filter_width: int,
    ):
        """
        Apply noise reduction to the spectrum.
        """
        if method == "box_filter":
            self.filtered, self.noise = box_filter_single(
                self.spectrum, filter_width
            )
            self.spectrum = self.filtered

    def continuum_removal(
        self,
        method: ContinuumMethodLiteral,
        tie_points: list[float] | None = None,
    ):
        if method == "double_line":
            self.contrem, self.continuum = double_line(
                self.spectrum, self.wavelength
            )
            self.spectrum = self.contrem
        elif method == "single_line":
            if tie_points is None:
                raise ValueError("Must specify tie points to use single_line.")
            self.contrem, self.continuum = single_line(
                self.spectrum, self.wavelength, tie_points=tie_points
            )
            self.spectrum = self.contrem

    def fit_absorption(
        self, low_wvl: float, high_wvl: float, unit: WvlUnit = "nm"
    ) -> AbsorptionFeature1D:
        if np.all(np.isnan(self.contrem)):
            raise ValueError("Continuum Removal has not been performed yet.")
        feature = AbsorptionFeature1D(
            self.contrem, self.wavelength, low_wvl, high_wvl, unit
        )
        return feature

    def _validate(self):
        if self.spectrum.ndim > 1:
            raise DimensionError(
                "Input spectrum array has too many dimensions "
                f"({self.spectrum.ndim})"
            )



# Dependencies
import numpy as np

# Relative Imports
from .utils import Interpolator

# Top-Level Imports
from reflspeckit.data_classes import Wavelength


def double_line(
    spectrum: np.ndarray, wvls: Wavelength
) -> tuple[np.ndarray, np.ndarray]:
    """
    Double-line continuum removal.

    Parameters
    ----------
    spectrum: np.ndarray
        Filtered spectrum with the continuum in place.
    wvls: Wavelength
        Wavelength values in the form of a Wavelength object.

    Returns
    -------
    continuum_removed: np.ndarray
        Input spectrum with the continuum removed.
    continuum: np.ndarray
        Continuum values for the spectrum.
    """
    # Getting initial continuum line parameters
    anchor_pts = np.array([700, 1550, 2600])
    if wvls.unit == "nm":
        pass
    if wvls.unit == "um":
        anchor_pts = anchor_pts * 10**-3
    if wvls.unit == "m":
        anchor_pts = anchor_pts * 10**-9
    if (wvls.values.min() > anchor_pts[0]) or (
        wvls.values.max() < anchor_pts[-1]
    ):
        raise ValueError(
            "Wavelength values are not in the correct range to perform lunar"
            "double line removal."
        )

    cont1_band_idx = np.empty(anchor_pts.size, dtype=np.int16)
    cont1_band_wvls = np.empty(anchor_pts.size, dtype=np.float64)
    for n in range(anchor_pts.size):
        idx = np.argmin(np.abs(wvls - anchor_pts[n]))
        cont1_band_idx[n] = idx
        cont1_band_wvls[n] = wvls.values[idx]

    cont1_spectrum_values = spectrum[cont1_band_idx]

    CI1 = Interpolator(cont1_band_wvls, cont1_spectrum_values)
    continuum1 = CI1.linear(wvls.values)

    continuum1_removed = spectrum / continuum1

    cont2_ranges: dict[str, tuple[float, float]] = {
        "range1": (650.0, 1000.0),
        "range2": (1250.0, 1600.0),
        "range3": (2000.0, 2800.0),
    }

    def scale_dict_values(
        data: dict[str, tuple[float, float]], factor: float
    ) -> dict[str, tuple[float, float]]:
        return {k: (v[0] * factor, v[1] * factor) for k, v in data.items()}

    if wvls.unit == "nm":
        pass
    elif wvls.unit == "um":
        cont2_ranges = scale_dict_values(cont2_ranges, 10**-3)
    elif wvls.unit == "m":
        cont2_ranges = scale_dict_values(cont2_ranges, 10**-9)
    else:
        raise ValueError("Wavelength Units not valid.")

    cont2_band_idx = np.empty(len(cont2_ranges), dtype=np.int16)

    # Locating dynamic maxima for second continuum parameters
    for n, val in enumerate(cont2_ranges.values()):
        lo_idx = np.argmin(np.abs(wvls.values - val[0]))
        hi_idx = np.argmin(np.abs(wvls.values - val[1]))

        max_spectrum_idx = (
            np.argmax(
                continuum1_removed[np.arange(lo_idx, hi_idx, dtype=np.int16)]
            )
            + lo_idx
        )
        cont2_band_idx[n] = max_spectrum_idx

    cont2_band_wvls = wvls.values[cont2_band_idx]
    cont2_spectrum_values = spectrum[cont2_band_idx]

    CI2 = Interpolator(cont2_band_wvls, cont2_spectrum_values)
    continuum2 = CI2.linear(wvls.values)
    continuum2_removed = spectrum / continuum2

    return continuum2_removed, continuum2


def single_line(
    spectrum: np.ndarray, wvls: Wavelength, tie_points: list[float]
) -> tuple[np.ndarray, np.ndarray]:
    if not all(
        [
            (i < wvls.values.max()) and (i > wvls.values.min())
            for i in tie_points
        ]
    ):
        raise ValueError("Tie points are outside of spectrum wavelengths")

    continuum_idx = np.empty(len(tie_points), dtype=np.int16)
    continuum_wvls = np.empty(len(tie_points), dtype=np.float64)
    for n in range(len(tie_points)):
        idx = np.argmin(np.abs(wvls.values - tie_points[n]))
        continuum_idx[n] = idx
        continuum_wvls[n] = wvls.values[idx]

    continuum_values = spectrum[continuum_idx]

    continuum_interp = Interpolator(continuum_wvls, continuum_values)
    continuum = continuum_interp.linear(wvls.values)
    continuum_removed = spectrum / continuum

    return continuum_removed, continuum



# Standard Libraries
from enum import Enum, auto

# Dependencies
import numpy as np

# Top-Level Imports
from reflspeckit.data_classes import (
    Wavelength,
    WvlUnit,
    FilterMethodLiteral,
    ContinuumMethodLiteral,
)
from reflspeckit._errors import DimensionError
from reflspeckit.utils import find_wvl, make_rgb_composite

# Relative Imports
from .outlier_detection import remove_outliers
from .filtering import box_filter_cube
from .continuum_removal import double_line
from .absorption_feature3d import AbsorptionFeature3D


class ProcessingFlag(Enum):
    RAW = auto()
    OUTLIERS_REMOVED = auto()
    FILTERED = auto()
    CONTINUUM_REMOVED = auto()


class Spec3D:
    """
    Main class for handling 3-dimensional spectral image cubes (i.e. an image
    where each pixel corresponds to a full spectrum).

    Parameters
    ----------
    spec_arr: np.ndarray
    wvl_arr: np.ndarray
        Wavelength values of the spectrum. Must be a 1D array.
    unit: WvlUnit, optional
        Unit for the wavelength values. Default is "nm", but options include:

        - "nm": Nanometers
        - "um": Microns (micrometers)
        - "m": Meters

    Notes
    -----
    See Spec3D for an equivalent class that handles 3-dimensional spectral data
    image cubes.

    Image cubes must be smaller than memory to use the standard Spec3D class.
    For image cubes larger than memory, see `StreamingSpec3D`.
    """

    def __init__(
        self, spec_arr: np.ndarray, wvl_arr: np.ndarray, unit: WvlUnit = "nm"
    ):
        self.cube = spec_arr
        self.wavelength = Wavelength(wvl_arr, unit)
        self._albedo_idx, _ = find_wvl(self.wavelength, 1580, unit=unit)

        self._processing_flag = ProcessingFlag.RAW

        self._validate()

    def outlier_removal(self, sigma_threshold: float = 1.5):
        print("Removing Outliers...")
        self.cube = remove_outliers(self.cube, sigma_threshold)
        self._processing_flag = ProcessingFlag.OUTLIERS_REMOVED

    def noise_reduction(self, method: FilterMethodLiteral, filter_width: int):
        print("Running Noise Reduction...")
        if self._processing_flag == ProcessingFlag.FILTERED:
            return
        if self._processing_flag != ProcessingFlag.OUTLIERS_REMOVED:
            self.outlier_removal()
        if method == "box_filter":
            self.cube, _ = box_filter_cube(self.cube, filter_width)
            self.albedo_image = self.cube[:, :, self._albedo_idx]

        self._processing_flag = ProcessingFlag.FILTERED

    def continuum_removal(self, method: ContinuumMethodLiteral):
        print("Removing Spectral Continuum...")
        if self._processing_flag == ProcessingFlag.CONTINUUM_REMOVED:
            return
        if self._processing_flag != ProcessingFlag.FILTERED:
            self.noise_reduction("box_filter", 7)
        if method == "double_line":
            self.cube, _ = double_line(self.cube, self.wavelength)

        self._processing_flag = ProcessingFlag.CONTINUUM_REMOVED

    def fit_absorption(
        self, low_wvl: float, high_wvl: float, unit: WvlUnit = "nm"
    ) -> AbsorptionFeature3D:
        print(
            f"Fitting Absorption Feature from {low_wvl} to {high_wvl} "
            f"{self.wavelength.unit}"
        )
        if self._processing_flag != ProcessingFlag.CONTINUUM_REMOVED:
            raise ValueError("Continuum Removal has not been performed yet.")
        feature = AbsorptionFeature3D(
            self.cube, self.wavelength, low_wvl, high_wvl, unit
        )
        return feature

    def make_m3_rgb(self):
        if self._processing_flag != ProcessingFlag.CONTINUUM_REMOVED:
            raise ValueError("Continuum Removal has not been performed yet.")
        bnd1 = self.fit_absorption(0.789, 1.309, unit="um")
        bnd2 = self.fit_absorption(1.658, 2.498, unit="um")

        ibd1 = bnd1.calculate_ibd()
        ibd2 = bnd2.calculate_ibd()

        rgb = make_rgb_composite(ibd1, ibd2, self.albedo_image)

        return rgb

    def _validate(self):
        if self.cube.ndim < 3:
            raise DimensionError(
                "Input spectrum array does not have enough dimensions "
                f"({self.cube.ndim})"
            )
        if self.cube.ndim > 3:
            raise DimensionError(
                "Input spectrum array has too many dimensions "
                f"({self.cube.ndim})"
            )



# utils/norm_image_controlled.py

# Standard Libraries
from copy import copy
from typing import Optional

# External Libraries
import numpy as np


def normalize_image(
    image_data: np.ndarray,
    low_threshold: Optional[float] = None,
    high_threshold: Optional[float] = None,
    min_val: float = 0,
    max_val: float = 1,
):
    """
    Normalizes an image from `min_val` to `max_val`, while cutting off image
    values below `low_threshold` and above `high_threshold` so that any value
    <`low`=`min_val` and any value >`high`=`max_val`.
    """
    norm_img = copy(image_data)
    no_nans = norm_img[np.isfinite(norm_img)]

    if (low_threshold is None) or (high_threshold is None):
        low = no_nans.min()
        high = no_nans.max()
    else:
        low = low_threshold
        high = high_threshold

    norm_img[norm_img > high] = high
    norm_img[norm_img < low] = low
    norm_img = (norm_img - low) / (high - low)
    norm_img = min_val + (norm_img * (max_val - min_val))
    return norm_img



# Standard Libraries
from typing import Callable

# Dependencies
import numpy as np


def round_to_odd(num):
    """
    Rounds a number to the nearest odd integer.
    """
    r = round(num, 0)
    if r % 2 == 0:
        if (num - r) != 0:
            return int(r + ((num - r) / abs(num - r)))
        else:
            return int(r - 1)
    else:
        return int(r)


class Interpolator:
    def __init__(self, xpts: np.ndarray, ypts: np.ndarray):
        """
        Parameters
        ----------
        xpts: np.ndarray
            1-D Array defining the tie points over which to interpolate.
        ypts: np.ndarray
            1-D Array representing the y points corresponding to the x points.
        """
        self.xpts = xpts
        self.ypts = ypts

    def linear(self, xvals: np.ndarray) -> np.ndarray:
        """
        Runs a linear interpolation over x values.
        """
        funclist: list[Callable] = []
        nseg = len(self.xpts) - 1
        for n in range(nseg):
            x1 = float(self.xpts[n])
            x2 = float(self.xpts[n + 1])
            y1 = self.ypts[n]
            y2 = self.ypts[n + 1]
            m = (y2 - y1) / (x2 - x1)
            b = y1 - m * x1

            funclist.append(lambda x, m=m, b=b: m * x + b)

        funclist.insert(0, funclist[0])
        funclist.append(funclist[-1])

        condlist = [xvals < self.xpts[0]]
        for n in range(nseg):
            if n == nseg - 1:
                cond = (xvals >= self.xpts[n]) & (xvals <= self.xpts[n + 1])
            else:
                cond = (xvals >= self.xpts[n]) & (xvals < self.xpts[n + 1])
            condlist.append(cond)
        condlist.append(xvals > self.xpts[-1])

        result = np.full(xvals.size, np.nan)
        for cnd, fnc in zip(condlist, funclist):
            result[cnd] = fnc(xvals[cnd])
        return result



# Standard Libraries
from typing import Callable

# Dependencies
import numpy as np


def round_to_odd(num):
    """
    Rounds a number to the nearest odd integer.
    """
    r = round(num, 0)
    if r % 2 == 0:
        if (num - r) != 0:
            return int(r + ((num - r) / abs(num - r)))
        else:
            return int(r - 1)
    else:
        return int(r)


class CubeInterpolator:
    def __init__(self, xpts: np.ndarray, ypts: np.ndarray):
        """
        Parameters
        ----------
        xpts: np.ndarray
            1-D Array defining the shared x points over which to interpolate
            the cube.
        ypts: np.ndarray
            3-D Array representing where the last axis contains the y points
            corresponding to the x points.
        """
        self.xpts = xpts
        self.ypts = ypts

    def linear(self, xvals: np.ndarray) -> np.ndarray:
        """
        Runs a linear interpolation over x values.
        """
        if self.xpts.ndim == 1:
            return self._linear_constX(xvals)
        elif self.xpts.ndim == 3:
            return self._linear_varX(xvals)
        else:
            raise ValueError(f"xvals is not valid. ({xvals.ndim} dims)")

    def _linear_constX(self, xvals: np.ndarray) -> np.ndarray:
        """
        Linear interpolation assuming that all pixels in a cube have the same
        tie points on the X axis. This method is called via `linear`.
        """
        # Build per-segment linear functions. Bind m and b into the lambda
        # default arguments to avoid late-binding closures.
        funclist: list[Callable] = []
        nseg = len(self.xpts) - 1
        for n in range(nseg):
            x1 = float(self.xpts[n])
            x2 = float(self.xpts[n + 1])
            y1 = self.ypts[:, :, n]
            y2 = self.ypts[:, :, n + 1]
            m = (y2 - y1) / (x2 - x1)
            b = y1 - m * x1

            # bind m and b so each function keeps its own slope/intercept
            funclist.append(
                lambda x, m=m, b=b: m[:, :, None] * x[None, None, :]
                + b[:, :, None]
            )

        funclist.insert(0, funclist[0])
        funclist.append(funclist[-1])

        # Create boolean masks per segment. Use left-inclusive, right-exclusive
        # bins except for the last segment which includes the right edge.
        condlist = [xvals < self.xpts[0]]
        for n in range(nseg):
            if n == nseg - 1:
                cond = (xvals >= self.xpts[n]) & (xvals <= self.xpts[n + 1])
            else:
                cond = (xvals >= self.xpts[n]) & (xvals < self.xpts[n + 1])
            condlist.append(cond)
        condlist.append(xvals > self.xpts[-1])

        result = np.full((*self.ypts.shape[:2], len(xvals)), np.nan)
        for cnd, fnc in zip(condlist, funclist):
            result[:, :, cnd] = fnc(xvals[cnd])
        return result

    def _linear_varX(self, xvals: np.ndarray) -> np.ndarray:
        nseg = self.xpts.shape[-1]
        slope_arr = np.empty((*self.ypts.shape[:2], self.xpts.shape[-1] - 1))
        intcp_arr = np.empty_like(slope_arr)
        for n in range(nseg - 1):
            x1 = self.xpts[:, :, n]
            x2 = self.xpts[:, :, n + 1]
            y1 = self.ypts[:, :, n]
            y2 = self.ypts[:, :, n + 1]
            slope_arr[:, :, n] = (y2 - y1) / (x2 - x1)
            intcp_arr[:, :, n] = y1 - slope_arr[:, :, n] * x1

        slope_arr = np.concat(
            [slope_arr, slope_arr[:, :, -1][:, :, None]], axis=2
        )
        intcp_arr = np.concat(
            [intcp_arr, intcp_arr[:, :, -1][:, :, None]], axis=2
        )

        eqn_arr = np.full((*self.ypts.shape[:2], xvals.size, 2), np.nan)
        for n in range(xvals.size):
            slope_arr_temp = slope_arr[:, :, 0]
            intcp_arr_temp = intcp_arr[:, :, 0]
            for j in range(self.xpts.shape[-1]):
                cond_arr_temp = xvals[n] > self.xpts[:, :, j]
                slope_arr_temp[cond_arr_temp] = slope_arr[cond_arr_temp, j]
                intcp_arr_temp[cond_arr_temp] = intcp_arr[cond_arr_temp, j]
            eqn_arr[:, :, n, 0] = slope_arr_temp
            eqn_arr[:, :, n, 1] = intcp_arr_temp

        line_arr = np.full((*self.ypts.shape[:2], xvals.size), np.nan)
        for n in range(xvals.size):
            line_arr[:, :, n] = (
                eqn_arr[:, :, n, 0] * xvals[n] + eqn_arr[:, :, n, 1]
            )
        return line_arr



# Dependencies
import numpy as np
import numpy.typing as npt

# Relative Imports
from .polyfit import polyfit_cube

# Top-Level Imports
from reflspeckit.data_classes import Wavelength, WvlUnit
from reflspeckit.utils import find_wvl
from reflspeckit._errors import WavelengthUnitError


class AbsorptionFeature3D:
    def __init__(
        self,
        contrem_spectrum: npt.NDArray,
        wavelength: Wavelength,
        low_wavelength: float,
        high_wavelength: float,
        unit: WvlUnit = "nm",
        fit_order: int = 4,
    ):
        self.spec = contrem_spectrum
        self.wvl = wavelength
        self.lowvl = low_wavelength
        self.hiwvl = high_wavelength

        if unit != self.wvl.unit:
            raise WavelengthUnitError(
                f"Absorption fit wavelengths are in units of {unit} whereas "
                f"the provided wavelegnth array is in units of {self.wvl.unit}"
            )

        low_idx, _lowwvl_exact = find_wvl(self.wvl.values, low_wavelength)
        high_idx, _highwvl_exact = find_wvl(self.wvl.values, high_wavelength)
        fit_slice = slice(low_idx, high_idx)

        self.fit_wvl = self.wvl.values[fit_slice]
        self._fit_spec = self.spec[:, :, fit_slice]

        self.fit_result = polyfit_cube(
            self.spec[:, :, fit_slice], self.wvl.values[fit_slice], fit_order
        )

    def calculate_center(self) -> npt.NDArray:
        center_idx_arr = np.argmin(self.fit_result.model, axis=2)
        bc_map = self.fit_wvl[center_idx_arr]
        bc_map[center_idx_arr == 0] = np.nan
        bc_map[center_idx_arr == self.spec.shape[-1] - 1] = np.nan

        return bc_map

    def calculate_ibd(self) -> npt.NDArray:
        return np.sum(1 - self._fit_spec, axis=2)



# Standard Libraries
from typing import Literal, Optional, TypeAlias

# Dependencies
import numpy as np

# Top-Level Imports
from reflspeckit._errors import DimensionError

# Wavelength unit type alias
WvlUnit = Literal["nm", "um", "m"]


class Wavelength:
    def __init__(
        self,
        values: np.ndarray,
        unit: WvlUnit,
        resolution: Optional[float] = None,
    ):
        self.values = values
        self.unit = unit
        self.resolution = resolution
        self._validate()

    def __array__(self):
        return self.values

    def to_nm(self):
        if self.unit == "nm":
            pass
        elif self.unit == "um":
            self.values *= 10**3
        elif self.unit == "m":
            self.values *= 10**9

        self.unit = "nm"

    def to_um(self):
        if self.unit == "nm":
            self.values *= 10**-3
        elif self.unit == "um":
            pass
        elif self.unit == "m":
            self.values *= 10**6

        self.unit = "um"

    def to_m(self):
        if self.unit == "nm":
            self.values *= 10**-9
        elif self.unit == "um":
            self.values *= 10**-6
        elif self.unit == "m":
            pass

        self.unit = "m"

    def _validate(self):
        if self.values.ndim > 1:
            raise DimensionError(
                "Wavelength array has too many dimensions "
                f"({self.values.ndim})"
            )


FilterMethodLiteral: TypeAlias = Literal["box_filter"]

ContinuumMethodLiteral: TypeAlias = Literal["double_line", "single_line"]



# Dependencies
import numpy as np
import numpy.typing as npt

# Relative Imports
from .polyfit import polyfit_single

# Top-Level Imports
from reflspeckit.data_classes import Wavelength, WvlUnit
from reflspeckit.utils import find_wvl
from reflspeckit._errors import WavelengthUnitError


class AbsorptionFeature1D:
    """
    Stores band parameter values for a specific absorption feature within a
    spectrum.

    Parameters
    ----------
    spectrum: npt.NDArray
        1-dimensional Array of continuum-removed spectrum values.
    low_wavelength: float
        Low end of absorption feature search.
    high_wavelength: float
        High end of absorption feature search.

    Attributes
    ----------

    Notes
    -----
    """

    def __init__(
        self,
        contrem_spectrum: npt.NDArray,
        wavelength: Wavelength,
        low_wavelength: float,
        high_wavelength: float,
        unit: WvlUnit = "nm",
        fit_order: int = 4,
    ) -> None:
        self.spec = contrem_spectrum
        self.wvl = wavelength
        self.lowvl = low_wavelength
        self.hiwvl = high_wavelength

        if unit != self.wvl.unit:
            raise WavelengthUnitError(
                f"Absorption fit wavelengths are in units of {unit} whereas "
                f"the provided wavelegnth array is in units of {self.wvl.unit}"
            )

        low_idx, _lowwvl_exact = find_wvl(self.wvl.values, low_wavelength)
        high_idx, _highwvl_exact = find_wvl(self.wvl.values, high_wavelength)

        fit_slice = slice(low_idx, high_idx)

        self.fit_wvl = self.wvl.values[fit_slice]
        self._fit_spec = self.spec[fit_slice]

        self.fit_result = polyfit_single(
            self.spec[fit_slice], self.wvl.values[fit_slice], fit_order
        )

    def calculate_center(self) -> tuple[float, float]:
        center_idx = np.argmin(self.fit_result.model)
        if (center_idx == 0) or (center_idx == self.spec.size - 1):
            return np.nan, np.nan
        else:
            return self.fit_result.ydata[center_idx], self.fit_wvl[center_idx]

    def calculate_ibd(self) -> float:
        return np.sum(1 - self._fit_spec)



# Dependencies
import numpy as np

# Top-Level Imports
from reflspeckit.data_classes import Wavelength, WvlUnit
from reflspeckit._errors import WavelengthUnitError


def find_wvl(
    wvls_input: Wavelength | np.ndarray, targetwvl: float, unit: WvlUnit = "nm"
):
    """
        findλ(λ.targetλ)

    Given a list of wavelengths, `wvls`, find the index of a `targetwvl` and
    the actual wavelength closest to your target.

    Parameters
    ----------
    wvls: np.ndarray
        Wavelength array to search in.
    targetwvl:
        Wavelength to search for.

    Returns
    -------
    idx: int
        Index of the found wavelength.
    wvl: float
        Actual wavelength that is closest to the target wavelength (at idx).
    """
    if isinstance(wvls_input, Wavelength):
        if wvls_input.unit != unit:
            raise WavelengthUnitError(
                f"{unit} does not match {wvls_input.unit}"
            )
        else:
            wvls = wvls_input.values
    else:
        wvls = wvls_input

    idx = np.argmin(np.abs(wvls - targetwvl))
    return idx, wvls[idx]



# Dependencies
import numpy as np

# Relative Imports
from .utils import round_to_odd
from .filtering import box_filter_single


def remove_outliers(original_spectrum: np.ndarray, threshold: float):
    """
    Detects and replaces statistical outliers in a 1D spectrum using
    neighbor-based interpolation.

    An outlier is defined as any value that deviates from a local moving
    average by more than a specified number of standard deviations
    (default is 2). Identified outliers are replaced by the mean of
    their immediate neighbors (i.e., the values before and after),
    excluding the outlier itself. Edge cases are handled by treating
    missing neighbors as NaN.

    Parameters
    ----------
    original_spectrum : np.ndarray
        The 1D input spectrum array to process. Must be a numeric NumPy array.
    threshold : float, optional
        The Z-score threshold to use for detecting outliers. Any value with a
        Z-score greater than `threshold` (in absolute value) is considered
        an outlier. Default is 2.

    Returns
    -------
    np.ndarray
        A new spectrum array with outliers replaced by the mean of their
        immediate neighbors. The original input is not modified.

    Notes
    -----
    - The local mean and standard deviation are computed using a moving
      window that spans 10% of the spectrum length (minimum size of 3
      and always rounded to an odd number).
    - To avoid circular wraparound, the first and last elements only
      use their single available neighbor for replacement.
    - This method is robust to isolated spikes, but may not perform well
      for broad or clustered outliers.

    Examples
    --------
    >>> import numpy as np
    >>> spectrum = np.array([1.0, 1.1, 1.2, 10.0, 1.3, 1.2, 1.1])
    >>> outlier_removal(spectrum)
    array([1. , 1.1, 1.2, 1.25, 1.3, 1.2, 1.1])
    """
    spec = np.copy(original_spectrum)
    window_size = max(round_to_odd(spec.size * 0.1), 3)
    mu, sig = box_filter_single(spec, window_size=window_size)
    zscore = (spec - mu) / sig
    outlier_idx = abs(zscore) > threshold

    neighbors = np.stack([np.roll(spec, -1), np.roll(spec, 1)], axis=1)

    neighbors[0, 1] = np.nan
    neighbors[-1, 0] = np.nan

    replacement = np.nanmean(neighbors, axis=1)

    spec[outlier_idx] = replacement[outlier_idx]

    return spec



# Dependencies
import numpy as np
import numpy.typing as npt
from scipy.ndimage import convolve1d

# Relative Imports
from .polyfit import polyfit_cube


def box_filter_cube(
    raw_cube: np.ndarray, window_size: int = 5
) -> tuple[npt.NDArray, npt.NDArray]:
    """
    Smoothes a spectrum using a moving average.

    Parameters
    ----------
    original_spectrum: np.ndarray
        Non-smooth spectral cube.
    window_size: optional, int
        Window size to use for the moving average. Default is 5.
    edge_handling: optional, str
        How to handle the edge cases of spectrum after convolution. Default is
        `"extrapolate"`
    remove_outliers: optional, bool
        If true, outliers are removed from the spectrum. A threshold value
        can be passed. Default is False.
    outlier_threshold: optional, float
        Sigma threshold to be used in outlier removal. See
        `outlier_removal.py`. Default is 2.
    """
    cube: npt.NDArray = raw_cube

    window = np.ones(window_size)
    nbands = cube.shape[2]

    edge_length = int(np.maximum(round(cube.shape[2] * 0.1, 0), 2))

    left_idx: npt.NDArray[np.int32] = np.arange(
        0, 1 + edge_length, dtype=np.int32
    )
    right_idx: npt.NDArray[np.int32] = np.arange(
        nbands - edge_length, nbands, dtype=np.int32
    )

    left_fit = polyfit_cube(cube[:, :, left_idx], left_idx, 1)
    left_ext = np.arange(-window_size, 0)
    left_arm = np.einsum(
        "ij,...j->...i", np.vander(left_ext, 2), left_fit.beta
    )

    right_fit = polyfit_cube(cube[:, :, right_idx], right_idx, 1)
    right_ext = np.arange(nbands, nbands + window_size)
    right_arm = np.einsum(
        "ij,...j->...i", np.vander(right_ext, 2), right_fit.beta
    )

    smoothed_spectrum = np.empty(
        (*cube.shape[0:2], nbands + left_arm.shape[-1] + right_arm.shape[-1])
    )

    smoothed_spectrum[:, :, : left_arm.shape[-1]] = left_arm
    smoothed_spectrum[
        :, :, left_arm.shape[-1] : -right_arm.shape[-1]  # noqa
    ] = cube
    smoothed_spectrum[:, :, -right_arm.shape[-1] :] = right_arm  # noqa

    mu = convolve1d(smoothed_spectrum, window, axis=2) / window_size
    musq = convolve1d(smoothed_spectrum**2, window, axis=2) / window_size
    sigma = np.sqrt(musq - mu**2)

    mu = mu[:, :, window_size:-window_size]
    sigma = sigma[:, :, window_size:-window_size]

    return mu, sigma
